import numpy as np
import mne
from scipy import signal
import matplotlib.pyplot as plt
from .constants import EEG_BANDS

class EEGSeries():

    def __init__(self, data: np.ndarray = None, path: str = None, sample_rate: int = 500) -> None:
        """Class for EEG time series.

        :param data: EEG data.
        :type data: np.ndarray
        :param path: Path to .npy array. Expected shape is (n_subjects, n_channels, n_samples)
        :type path: str
        :param sample_rate: sample rate, defaults to 500
        :type sample_rate: int, optional
        """
        assert(data is not None or path is not None)
        self.data = np.load(path) if path is not None else data
        assert(len(self.data.shape) == 3)
        self.sample_rate = sample_rate

    def apply_cheby_filter(self, lowcut: float, highcut: float, order: int=6, rs: float=40, plot_response=False):
        """Apply a Chebyshev II filter to the EEG data.

        :param lowcut: Lower pass-band edge.
        :type lowcut: float
        :param highcut: Upper pass-band edge.
        :type highcut: float
        :param order: [description], defaults to 6
        :type order: int, optional
        :param rs: [description], defaults to 40
        :type rs: float, optional
        :param plot_response: [description], defaults to False
        :type plot_response: bool, optional
        :return: Filtered data
        :rtype: EEGSeries
        """


        nyq = 0.5 * self.sample_rate
        low = lowcut / nyq
        high = highcut / nyq
        sos = signal.cheby2(order, rs, [low, high], btype='band', output='sos')
        if plot_response:
            w, h = signal.sosfreqz(sos)
            plt.plot((self.sample_rate * 0.5 / np.pi)
                     * w, 20 * np.log10(abs(h)))
            plt.title('Frequency response (rs={})'.format(rs))
            plt.xlabel('Frequency')
            plt.ylabel('Amplitude [dB]')
            plt.margins(0, 0.1)
            plt.grid(which='both', axis='both')
            plt.axvline(lowcut, color='red')
            plt.axvline(highcut, color='red')
            plt.show()
        return EEGSeries(data=signal.sosfilt(sos, self.data))

    def fir_filter(self, l_freq: float, h_freq: float,  verbose=False, **kwargs) -> np.ndarray:
        """Apply a FIR filter to the EEG data. Accepts arguments for mne.filter.filter_data.

        :param l_freq: Lower pass-band edge.
        :type l_freq: float
        :param h_freq: Upper pass-band edge.
        :type h_freq: float
        :return: Filtered data
        :rtype: EEGSeries
        """
        return EEGSeries(data=mne.filter.filter_data(self.data, self.sample_rate, l_freq, h_freq, verbose=verbose, **kwargs))

    def fir_filter_bands(self, **kwargs) -> dict:
        """Return a dictionary of filtered EEG data. 

        :return: Dictionary of filtered data. {band_name: data}
        :rtype: dict
        """
        return {band_name: self.fir_filter(band_range[0], band_range[1], **kwargs) for band_name, band_range in EEG_BANDS.items()}

    def cheby_filter_bands(self, **kwargs) -> dict:
        """Return a dictionary of filtered EEG data. 

        :return: Dictionary of filtered data. {band_name: data}
        :rtype: dict
        """
        return {band_name: self.apply_cheby_filter(band_range[0], band_range[1], **kwargs) for band_name, band_range in EEG_BANDS.items()}

    def append(self, new_data) -> None:
        """Append new data to the EEG data.

        :param new_data: Data to append.
        :type new_data: EEGSeries
        """
        return EEGSeries(data = np.append(self.data, new_data.data, axis=0))

    def __iter__(self):
        return iter(self.data)





from .timeseries import EEGSeries
import numpy as np


def window_data_loader(eegs: EEGSeries, batch_size: int = 32, window_size: float = 1, infinity: bool = False, labels: np.ndarray = None, epochs=1, shuffle: bool = False, return_subjects=False, stride: float = None) -> tuple:
    """[summary]

    :param eegs: [description]
    :type eegs: EEGSeries
    :param batch_size: batch size, defaults to 32
    :type batch_size: int, optional
    :param window_size: eeg window in seconds, defaults to 1
    :type window_size: float, optional
    :param infinity: whether the loop should be infinity, defaults to False
    :type infinity: bool, optional
    :param labels: Labels per subject, defaults to None
    :type labels: np.ndarray, optional
    :param epochs: number of epochs, defaults to 1
    :type epochs: int, optional
    :param shuffle: whether to shuffle the data, defaults to False
    :type shuffle: bool, optional
    :param return_subjects: batch size correspond to number of windows per subject, defaults to False
    :type return_subjects: bool, optional
    :param stride: stride in seconds, defaults to None
    :type stride: float, optional
    :yield: batch and labels if labels are provided
    :rtype: tuple
    """


    window_sample = int(window_size * eegs.sample_rate)
    data = np.lib.stride_tricks.sliding_window_view(eegs.data, window_sample, axis=-1)
    if stride is not None:
        data = data[:,:,:: int(eegs.sample_rate * stride),:] 
    else:
        data = data[:,:,::window_sample,:]
    data = data.reshape(-1, data.shape[1], data.shape[-1])

    if return_subjects:
        batch_size = data.shape[0] // eegs.data.shape[0]

    if labels is not None:
        new_labels = np.repeat(labels, data.shape[0] // eegs.data.shape[0])

    batch_number = 0
    for epoch in range(epochs):
        done_epoch = None

        if shuffle:
            p = np.random.permutation(len(data))
            data = data[p]
            if labels is not None:
                new_labels = new_labels[p]

        while done_epoch is None:
            if batch_number * batch_size >= data.shape[0]:
                batch_number = 0
                done_epoch = epoch
                data_batch = data[batch_number * batch_size:]
                if labels is not None:
                    label_batch = new_labels[batch_number * batch_size:]

            data_batch = data[batch_number *
                              batch_size:(batch_number + 1) * batch_size]
            if labels is not None:
                label_batch = new_labels[batch_number *
                                         batch_size:(batch_number + 1) * batch_size]
            batch_number += 1
            yield data_batch, label_batch if labels is not None else None, done_epoch
